//---------------------------------------------------------------------------
#include <vcl.h>
#include <vcl\inifiles.hpp>
#include <mmsystem.h>
#include <stdio.h>
#include <math.h>
#pragma hdrstop

#include "rtklib.h"
#include "strdlg.h"
#include "mondlg.h"
#include "tcpoptdlg.h"
#include "confdlg.h"
#include "aboutdlg.h"
#include "viewer.h"
#include "naviopt.h"
#include "navimain.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainForm *MainForm;

#define PRGNAME     "RTKNAVI"		// program name
#define TRACEFILE	"rtknavi.trace" // debug trace file
#define CLORANGE    (TColor)0x00AAFF
#define CLLGRAY     (TColor)0xDDDDDD
#define CHARDEG		"\260"			// character code of degree
#define SATSIZE		18				// satellite cycle size in skyplot
#define SOLSIZE		10				// solution cycle size in track
#define HISSIZE		4				// history cycle size in track
#define VELSIZE		40				// velocity cycle size in track
#define MINSNR		10				// minimum snr
#define MAXSNR		60				// maximum snr
#define KEYF6		0x75			// code of function key F6
#define KEYF7		0x76			// code of function key F7
#define KEYF8		0x77			// code of function key F8
#define KEYF9		0x78			// code of function key F9
#define KEYF10		0x79			// code of function key F10
#define POSFONTNAME	"Palatino Linotype"
#define POSFONTSIZE	12

#define SQRT(x)		((x)<0.0?0.0:sqrt(x))

//---------------------------------------------------------------------------

rtksvr_t rtksvr;					// rtk server struct

//---------------------------------------------------------------------------
extern "C" {
extern int showmsg(char *format,...) {return 0;}
}
//---------------------------------------------------------------------------
static void degtodms(double deg, double *dms)
{
    double sgn=1.0;
    
    if (deg<0.0) {deg=-deg; sgn=-1.0;}
    dms[0]=floor(deg);
    dms[1]=floor((deg-dms[0])*60.0);
    dms[2]=(deg-dms[0]-dms[1]/60.0)*3600;
    dms[0]*=sgn;
}
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
	: TForm(Owner)
{
	SvrCycle=SvrBuffSize=0; Scale=6;
	SolBuffSize=1000;
	for (int i=0;i<MAXSTRRTK;i++) {
		Stream[i]=Format[i]=CmdEna[i][0]=CmdEna[i][1]=0;
	}
	TimeSys=SolType=PlotType=0;
	PSol=PSolS=PSolE=Nsat=0;
	Time=NULL;
	SolStat=Nvsat=NULL;
	SolRov=SolRef=Qr=VelRov=Age=Ratio=NULL;
	for (int i=0;i<MAXSAT;i++) {
		Sat[i]=Snr1[i]=Snr2[i]=Vsat[i]=0;
		Az[i]=El[i]=0.0;
	}
	PrcOpt=prcopt_default;
	SolOpt=solopt_default;
	PosFont=new TFont;
	
	rtksvrinit(&rtksvr);
	
	Caption=PRGNAME;
	Caption=Caption+" ver."+VER_RTKLIB;
	DoubleBuffered=true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormShow(TObject *Sender)
{
	TIniFile *ini=new TIniFile("rtknavi.ini");
	AnsiString s;
	
	for (int i=0;i<MAXSTRRTK;i++) {
		Stream[i]=ini->ReadInteger("stream",s.sprintf("stream%d",i),0);
		Format[i]=ini->ReadInteger("stream",s.sprintf("format%d",i),0);
		for (int j=0;j<3;j++) {
			Paths[i][j]=ini->ReadString("stream",s.sprintf("path_%d_%d",i,j),"");
		}
	}
	for (int i=0;i<2;i++) for (int j=0;j<2;j++) {
		Cmds[i][j]=ini->ReadString("serial",s.sprintf("cmd_%d_%d",i,j),"");
		CmdEna[i][j]=ini->ReadInteger("serial",s.sprintf("cmdena_%d_%d",i,j),0);
		for (char *p=Cmds[i][j].c_str();*p;p++) {
			if ((p=strstr(p,"@@"))) strncpy(p,"\r\n",2); else break;
		}
	}
	PrcOpt.mode  =ini->ReadInteger("prcopt","mode",2);
	PrcOpt.nf    =ini->ReadInteger("prcopt","nf",2);
	PrcOpt.elmin =ini->ReadFloat("prcopt","elmin",15.0*D2R);
	PrcOpt.snrmin=ini->ReadFloat("prcopt","snrmin",0.0);
	PrcOpt.intpref=ini->ReadInteger("prcopt","intpref",1);
	PrcOpt.modear=ini->ReadInteger("prcopt","modear",0.0);
	PrcOpt.maxout=ini->ReadInteger("prcopt","maxout",5);
	PrcOpt.minlock=ini->ReadInteger("prcopt","minlock",0);
	PrcOpt.estiono=ini->ReadInteger("prcopt","estiono",0);
	PrcOpt.esttrop=ini->ReadInteger("prcopt","esttrop",0);
	PrcOpt.niter  =ini->ReadInteger("prcopt","niter",1);
	PrcOpt.err[0] =ini->ReadFloat("prcopt","err0",100.0);
	PrcOpt.err[1] =ini->ReadFloat("prcopt","err1",0.003);
	PrcOpt.err[2] =ini->ReadFloat("prcopt","err2",0.003);
	PrcOpt.err[3] =ini->ReadFloat("prcopt","err3",0.0);
	PrcOpt.prn[0] =ini->ReadFloat("prcopt","prn0",1E-4);
	PrcOpt.prn[1] =ini->ReadFloat("prcopt","prn1",1E-3);
	PrcOpt.prn[2] =ini->ReadFloat("prcopt","prn2",1E-4);
	PrcOpt.sclkstab=ini->ReadFloat("prcopt","sclkstab",5E-12);
	PrcOpt.thresar=ini->ReadFloat("prcopt","thresar",3.0);
	PrcOpt.elmaskar=ini->ReadFloat("prcopt","elmaskar",0.0);
	PrcOpt.thresslip=ini->ReadFloat("prcopt","thresslip",0.05);
	PrcOpt.maxtdiff=ini->ReadFloat("prcopt","maxtdiff",30.0);
	PrcOpt.maxinno=ini->ReadFloat("prcopt","maxinno",30.0);
	PrcOpt.sbascorr=ini->ReadInteger("setting","sbascorr",0);
	
	SolOpt.posf  =ini->ReadInteger("solopt","posf",0);
	SolOpt.times =ini->ReadInteger("solopt","times",0);
	SolOpt.timef =ini->ReadInteger("solopt","timef",1);
	SolOpt.degf  =ini->ReadInteger("solopt","degf",0);
	s=ini->ReadString("solopt","sep"," ");
	strcpy(SolOpt.sep,s.c_str());
	SolOpt.outhead=ini->ReadInteger("solopt","outhead",0);
	SolOpt.outopt=ini->ReadInteger("solopt","outopt",0);
	SolOpt.datum=ini->ReadInteger("solopt","datum",0);
	SolOpt.height=ini->ReadInteger("solopt","height",0);
	SolOpt.geoid=ini->ReadInteger("solopt","geoid",0);
	
	TimeSys    =ini->ReadInteger("setting","timesys",0);
	SolType    =ini->ReadInteger("setting","soltype",0);
	SvrCycle   =ini->ReadInteger("setting","svrcycle",10);
	SvrBuffSize=ini->ReadInteger("setting","svrbuffsize",32768);
	SolBuffSize=ini->ReadInteger("setting","solbuffsize",1000);
	NavSelect  =ini->ReadInteger("setting","navselect",0);
	SbasSat    =ini->ReadInteger("setting","sbassat",0);
	SavedSol   =ini->ReadInteger("setting","savedsol",100);
	DebugTraceF=ini->ReadInteger("setting","debugtrace",0);
	RovPosTypeF=ini->ReadInteger("setting","rovpostype",0);
	RefPosTypeF=ini->ReadInteger("setting","refpostype",0);
	RovAntPcvF =ini->ReadInteger("setting","rovantpcv",0);
	RefAntPcvF =ini->ReadInteger("setting","refantpcv",0);
	RovAntF    =ini->ReadString("setting","rovant","");
	RefAntF    =ini->ReadString("setting","refant","");
	AntPcvFileF=ini->ReadString("setting","antpcvfile","");
	StaPosFileF=ini->ReadString("setting","staposfile","");
	GoogleEarthFileF=ini->ReadString("setting","googleearthfile","");
	for (int i=0;i<3;i++) {
		RovAntDel[i]=ini->ReadFloat("setting",s.sprintf("rovantdel_%d",i),0.0);
		RefAntDel[i]=ini->ReadFloat("setting",s.sprintf("refantdel_%d",i),0.0);
		RovPos[i]=ini->ReadFloat("setting",s.sprintf("rovpos_%d",i),0.0);
		RefPos[i]=ini->ReadFloat("setting",s.sprintf("refpos_%d",i),0.0);
	}
	for (int i=0;i<10;i++) {
		TcpOptDialog->History[i]=ini->ReadString("tcpopt",s.sprintf("history%d",i),"");
	}
	PosFont->Charset=ANSI_CHARSET;
	PosFont->Name=ini->ReadString("setting","posfontname",POSFONTNAME);
	PosFont->Size=ini->ReadInteger("setting","posfontsize",POSFONTSIZE);
	PosFont->Color=(TColor)ini->ReadInteger("setting","posfontcolor",(int)clBlack);
	if (ini->ReadInteger("setting","posfontbold",  0)) PosFont->Style=PosFont->Style<<fsBold;
	if (ini->ReadInteger("setting","posfontitalic",0)) PosFont->Style=PosFont->Style<<fsItalic;

	TTextViewer::Color1=(TColor)ini->ReadInteger("viewer","color1",(int)clBlack);
	TTextViewer::Color2=(TColor)ini->ReadInteger("viewer","color2",(int)clWhite);
	TTextViewer::FontD=new TFont;
	TTextViewer::FontD->Name=ini->ReadString("viewer","fontname","Courier New");
	TTextViewer::FontD->Size=ini->ReadInteger("viewer","fontsize",9);
	delete ini;
	
	LoadNav(&rtksvr.nav);
	InitSolBuff();
	
	UpdateTimeSys();
	UpdateSolType();
	UpdateFont();
	UpdatePos();
	UpdatePlot();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormClose(TObject *Sender, TCloseAction &Action)
{
	TIniFile *ini=new TIniFile("rtknavi.ini");
	AnsiString s;
	
	for (int i=0;i<MAXSTRRTK;i++) {
		ini->WriteInteger("stream",s.sprintf("stream%d",i),Stream[i]);
		ini->WriteInteger("stream",s.sprintf("format%d",i),Format[i]);
		for (int j=0;j<3;j++) {
			ini->WriteString("stream",s.sprintf("path_%d_%d",i,j),Paths[i][j]);
		}
	}
	for (int i=0;i<2;i++) for (int j=0;j<2;j++) {
		for (char *p=Cmds[i][j].c_str();*p;p++) {
			if ((p=strstr(p,"\r\n"))) strncpy(p,"@@",2); else break;
		}
		ini->WriteString("serial",s.sprintf("cmd_%d_%d",i,j),Cmds[i][j]);
		ini->WriteInteger("serial",s.sprintf("cmdena_%d_%d",i,j),CmdEna[i][j]);
	}
	ini->WriteInteger("prcopt","mode",PrcOpt.mode);
	ini->WriteInteger("prcopt","nf",PrcOpt.nf);
	ini->WriteFloat("prcopt","elmin",PrcOpt.elmin);
	ini->WriteFloat("prcopt","snrmin",PrcOpt.snrmin);
	ini->WriteInteger("prcopt","modear",PrcOpt.modear);
	ini->WriteInteger("prcopt","intpref",PrcOpt.intpref);
	ini->WriteInteger("prcopt","modear",PrcOpt.modear);
	ini->WriteInteger("prcopt","maxout",PrcOpt.maxout);
	ini->WriteInteger("prcopt","minlock",PrcOpt.minlock);
	ini->WriteInteger("prcopt","estiono",PrcOpt.estiono);
	ini->WriteInteger("prcopt","esttrop",PrcOpt.esttrop);
	ini->WriteInteger("prcopt","niter",PrcOpt.niter);
	ini->WriteFloat("prcopt","err0",PrcOpt.err[0]);
	ini->WriteFloat("prcopt","err1",PrcOpt.err[1]);
	ini->WriteFloat("prcopt","err2",PrcOpt.err[2]);
	ini->WriteFloat("prcopt","err3",PrcOpt.err[3]);
	ini->WriteFloat("prcopt","prn0",PrcOpt.prn[0]);
	ini->WriteFloat("prcopt","prn1",PrcOpt.prn[1]);
	ini->WriteFloat("prcopt","prn2",PrcOpt.prn[2]);
	ini->WriteFloat("prcopt","sclkstab",PrcOpt.sclkstab);
	ini->WriteFloat("prcopt","thresar",PrcOpt.thresar);
	ini->WriteFloat("prcopt","elmaskar",PrcOpt.elmaskar);
	ini->WriteFloat("prcopt","thresslip",PrcOpt.thresslip);
	ini->WriteFloat("prcopt","maxtdiff",PrcOpt.maxtdiff);
	ini->WriteFloat("prcopt","maxinno",PrcOpt.maxinno);
	ini->WriteInteger("setting","sbascorr",PrcOpt.sbascorr);
	
	ini->WriteInteger("solopt","posf",SolOpt.posf);
	ini->WriteInteger("solopt","times",SolOpt.times);
	ini->WriteInteger("solopt","timef",SolOpt.timef);
	ini->WriteInteger("solopt","degf",SolOpt.degf);
	ini->WriteString("solopt","sep",SolOpt.sep);
	ini->WriteInteger("solopt","outhead",SolOpt.outhead);
	ini->WriteInteger("solopt","outopt",SolOpt.outopt);
	ini->WriteInteger("solopt","datum",SolOpt.datum);
	ini->WriteInteger("solopt","height",SolOpt.height);
	ini->WriteInteger("solopt","geoid",SolOpt.geoid);
	
	ini->WriteInteger("setting","timesys",TimeSys);
	ini->WriteInteger("setting","soltype",SolType);
	ini->WriteInteger("setting","svrcycle",SvrCycle);
	ini->WriteInteger("setting","svrbuffsize",SvrBuffSize);
	ini->WriteInteger("setting","solbuffsize",SolBuffSize);
	ini->WriteInteger("setting","navselect",NavSelect);
	ini->WriteInteger("setting","sbassat",SbasSat);
	ini->WriteInteger("setting","savedsol",SavedSol);
	ini->WriteInteger("setting","debugtrace",DebugTraceF);
	ini->WriteInteger("setting","rovpostype",RovPosTypeF);
	ini->WriteInteger("setting","refpostype",RefPosTypeF);
	ini->WriteInteger("setting","rovantpcv",RovAntPcvF);
	ini->WriteInteger("setting","refantpcv",RefAntPcvF);
	ini->WriteString("setting","rovant",RovAntF);
	ini->WriteString("setting","refant",RefAntF);
	ini->WriteString("setting","antpcvfile",AntPcvFileF);
	ini->WriteString("setting","staposfile",StaPosFileF);
	ini->WriteString("setting","googleearthfile",GoogleEarthFileF);
	for (int i=0;i<3;i++) {
		ini->WriteFloat("setting",s.sprintf("rovantdel_%d",i),RovAntDel[i]);
		ini->WriteFloat("setting",s.sprintf("refantdel_%d",i),RefAntDel[i]);
		ini->WriteFloat("setting",s.sprintf("rovpos_%d",i),RovPos[i]);
		ini->WriteFloat("setting",s.sprintf("refpos_%d",i),RefPos[i]);
	}
	for (int i=0;i<10;i++) {
		if (TcpOptDialog->History[i]=="") continue;
		ini->WriteString("tcpopt",s.sprintf("history%d",i),TcpOptDialog->History[i]);
	}
	ini->WriteString ("setting","posfontname",PosFont->Name);
	ini->WriteInteger("setting","posfontsize",PosFont->Size);
	ini->WriteInteger("setting","posfontcolor",(int)PosFont->Color);
	ini->WriteInteger("setting","posfontbold",  PosFont->Style.Contains(fsBold));
	ini->WriteInteger("setting","posfontitalic",PosFont->Style.Contains(fsItalic));
	
	ini->WriteInteger("viewer","color1",(int)TTextViewer::Color1);
	ini->WriteInteger("viewer","color2",(int)TTextViewer::Color2);
	ini->WriteString ("viewer","fontname",TTextViewer::FontD->Name);
	ini->WriteInteger("viewer","fontsize",TTextViewer::FontD->Size);
	delete ini;
	
	SaveNav(&rtksvr.nav);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnExitClick(TObject *Sender)
{
	Close();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnStartClick(TObject *Sender)
{
	AnsiString s;
	int itype[]={STR_NONE,STR_SERIAL,STR_TCPCLI,STR_TCPSVR,STR_NTRIPCLI,STR_FILE};
	int otype[]={STR_NONE,STR_SERIAL,STR_TCPCLI,STR_TCPSVR,STR_NTRIPSVR,STR_FILE};
	int ip[]={0,1,1,1,2},strs[MAXSTRRTK]={0};
	char *paths[MAXSTRRTK],*cmds[2]={0};
	FILE *fp;
	
	Message->Caption="";
	
	if (DebugTraceF>0) {
		traceopen(TRACEFILE); tracelevel(DebugTraceF);
	}
	PrcOpt.rb[0]=RefPos[0];
	PrcOpt.rb[1]=RefPos[1];
	PrcOpt.rb[2]=RefPos[2];
	for (int i=0;i<2;i++) strs[i]=itype[Stream[i]];
	for (int i=2;i<4;i++) strs[i]=otype[Stream[i]];
	for (int i=4;i<6;i++) strs[i]=otype[Stream[i]];
	for (int i=0;i<MAXSTRRTK;i++) {
		paths[i]=Stream[i]==0?"":Paths[i][ip[Stream[i]-1]].c_str();
	}
	for (int i=0;i<2;i++) {
		if (CmdEna[i][0]) cmds[i]=Cmds[i][0].c_str();
	}
	for (int i=2;i<6;i++) {
		if (strs[i]!=STR_FILE||!(fp=fopen(paths[i],"r"))) continue;
		fclose(fp);
		for (int j=0;j<2;j++) {
			if (strs[j]==STR_FILE||!strcmp(paths[i],paths[j])) {
				Message->Caption=s.sprintf("invalid output (%d)",i+1);
				return;
			}
		}
		ConfDialog->Label2->Caption=paths[i];
		if (ConfDialog->ShowModal()!=mrOk) return;
	}
	if (!rtksvrstart(&rtksvr,SvrCycle,SvrBuffSize,strs,paths,Format,NavSelect,
					 SbasSat,cmds,&PrcOpt,&SolOpt)) return;
	
	PSol=PSolS=PSolE=0;
	SolStat[0]=Nvsat[0]=0;
	for (int i=0;i<3;i++) SolRov[i]=SolRef[i]=VelRov[i]=0.0;
	for (int i=0;i<9;i++) Qr[i]=0.0;
	Age[0]=Ratio[0]=0.0;
	UpdatePos();
	UpdatePlot();
	BtnStart->Enabled=false;
	BtnExit->Enabled=false;
	BtnStr->Enabled=false;
	BtnOpt->Enabled=false;
	BtnStop->Enabled=true;
	Svr->Color=CLORANGE;
	ScbSol->Enabled=false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnStopClick(TObject *Sender)
{
	char *cmds[2]={0};
	int n,m;
	
	for (int i=0;i<2;i++) {
		if (CmdEna[i][1]) cmds[i]=Cmds[i][1].c_str();
	}
	rtksvrstop(&rtksvr,cmds);
	
	BtnStart->Enabled=true;
	BtnExit->Enabled=true;
	BtnStr->Enabled=true;
	BtnOpt->Enabled=true;
	BtnStop->Enabled=false;
	LabelTime->Font->Color=clGray;
	Svr->Color=clBtnFace;
	IndSol->Color=clWhite;
	ScbSol->Enabled=true;
	n=PSolE-PSolS; if (n<0) n+=SolBuffSize;
	m=PSol-PSolS;  if (m<0) m+=SolBuffSize;
	if (n>0) {
		ScbSol->Max=n-1; ScbSol->Position=m;
	}
	if (DebugTraceF>0) traceclose();
	
	Message->Caption="";
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnOptClick(TObject *Sender)
{
	AnsiString s;
	OptDialog->PrcOpt=PrcOpt;
	OptDialog->SolOpt=SolOpt;
	OptDialog->DebugTraceF=DebugTraceF;
	OptDialog->RovPosTypeF=RovPosTypeF;
	OptDialog->RefPosTypeF=RefPosTypeF;
	OptDialog->RovAntPcvF=RovAntPcvF;
	OptDialog->RefAntPcvF=RefAntPcvF;
	OptDialog->RovAntF=RovAntF;
	OptDialog->RefAntF=RefAntF;
	OptDialog->AntPcvFileF=AntPcvFileF;
	OptDialog->StaPosFileF=StaPosFileF;
	OptDialog->GoogleEarthFileF=GoogleEarthFileF;
	OptDialog->SvrCycle=SvrCycle;
	OptDialog->SvrBuffSize=SvrBuffSize;
	OptDialog->SolBuffSize=SolBuffSize;
	OptDialog->NavSelect=NavSelect;
	OptDialog->SbasSat=SbasSat;
	OptDialog->SavedSol=SavedSol;
	for (int i=0;i<3;i++) {
		OptDialog->RovAntDel[i]=RovAntDel[i];
		OptDialog->RefAntDel[i]=RefAntDel[i];
		OptDialog->RovPos[i]=RovPos[i];
		OptDialog->RefPos[i]=RefPos[i];
	}
	OptDialog->PosFont->Assign(PosFont);
	
	if (OptDialog->ShowModal()!=mrOk) return;
	
	PrcOpt=OptDialog->PrcOpt;
	SolOpt=OptDialog->SolOpt;
	DebugTraceF=OptDialog->DebugTraceF;
	RovPosTypeF=OptDialog->RovPosTypeF;
	RefPosTypeF=OptDialog->RefPosTypeF;
	RovAntPcvF=OptDialog->RovAntPcvF;
	RefAntPcvF=OptDialog->RefAntPcvF;
	RovAntF=OptDialog->RovAntF;
	RefAntF=OptDialog->RefAntF;
	AntPcvFileF=OptDialog->AntPcvFileF;
	StaPosFileF=OptDialog->StaPosFileF;
	GoogleEarthFileF=OptDialog->GoogleEarthFileF;
	SvrCycle=OptDialog->SvrCycle;
	SvrBuffSize=OptDialog->SvrBuffSize;
	NavSelect=OptDialog->NavSelect;
	SbasSat=OptDialog->SbasSat;
	SavedSol=OptDialog->SavedSol;
	if (SolBuffSize!=OptDialog->SolBuffSize) {
		SolBuffSize=OptDialog->SolBuffSize;
		InitSolBuff();
		UpdateTime();
		UpdatePos();
		UpdatePlot();
	}
	for (int i=0;i<3;i++) {
		RovAntDel[i]=OptDialog->RovAntDel[i];
		RefAntDel[i]=OptDialog->RefAntDel[i];
		RovPos[i]=OptDialog->RovPos[i];
		RefPos[i]=OptDialog->RefPos[i];
	}
	PosFont->Assign(OptDialog->PosFont);
	if (PrcOpt.pcvu) free(PrcOpt.pcvu);
	PrcOpt.pcvu=NULL;
	if (RovAntPcvF) {
		if (!(PrcOpt.pcvu=(pcv_t *)malloc(sizeof(pcv_t)))) return;
		if (readpcv(AntPcvFileF.c_str(),RovAntF.c_str(),PrcOpt.pcvu)) {
			Message->Caption=s.sprintf("no antenna pcv %s",RovAntF.c_str());
			return;
		}
		for (int i=0;i<3;i++) PrcOpt.pcvu->ecc[i]=RovAntDel[i];
	}
	if (PrcOpt.pcvr) free(PrcOpt.pcvr);
	PrcOpt.pcvr=NULL;
	if (RefAntPcvF) {
		if (!(PrcOpt.pcvr=(pcv_t *)malloc(sizeof(pcv_t)))) return;
		if (readpcv(AntPcvFileF.c_str(),RefAntF.c_str(),PrcOpt.pcvr)) {
			Message->Caption=s.sprintf("no antenna pcv %s",RefAntF.c_str());
			return;
		}
		for (int i=0;i<3;i++) PrcOpt.pcvr->ecc[i]=RefAntDel[i];
	}
	UpdateFont();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnStrClick(TObject *Sender)
{
	for (int i=0;i<MAXSTRRTK;i++) {
		StrDialog->Stream[i]=Stream[i];
		StrDialog->Format[i]=Format[i];
		for (int j=0;j<3;j++) StrDialog->Paths[i][j]=Paths[i][j];
	}
	for (int i=0;i<2;i++) for (int j=0;j<2;j++) {
		StrDialog->CmdEna[i][j]=CmdEna[i][j];
		StrDialog->Cmds[i][j]=Cmds[i][j];
	}
	if (StrDialog->ShowModal()!=mrOk) return;
	for (int i=0;i<MAXSTRRTK;i++) {
		Stream[i]=StrDialog->Stream[i];
		Format[i]=StrDialog->Format[i];
		for (int j=0;j<3;j++) Paths[i][j]=StrDialog->Paths[i][j];
	}
	for (int i=0;i<2;i++) for (int j=0;j<2;j++) {
		CmdEna[i][j]=StrDialog->CmdEna[i][j];
		Cmds[i][j]=StrDialog->Cmds[i][j];
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnTimeSysClick(TObject *Sender)
{
	if (++TimeSys>3) TimeSys=0;
	UpdateTimeSys();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnSolTypeClick(TObject *Sender)
{
	if (++SolType>4) SolType=0;
	UpdateSolType();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnPlotTypeClick(TObject *Sender)
{
	ChangePlot(1);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnMonitorClick(TObject *Sender)
{
	MonitorDialog->Show();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ScbSolChange(TObject *Sender)
{
	PSol=PSolS+ScbSol->Position;
	if (PSol>=SolBuffSize) PSol-=SolBuffSize;
	UpdateTime();
	UpdatePos();
	if (PlotType>=4) UpdatePlot();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnSaveClick(TObject *Sender)
{
	SaveLog();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UdScaleChangingEx(TObject *Sender, bool &AllowChange,
      short NewValue, TUpDownDirection Direction)
{
	if (Direction==updUp) ExpandPlot(1);
	else if (Direction==updDown) ExpandPlot(0);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	switch (Key) {
		case KEYF6 : SaveLog();     break;
		case KEYF7 : ExpandPlot(1); break;
		case KEYF8 : ExpandPlot(0); break;
		case KEYF9 : ChangePlot(0); break;
		case KEYF10: ChangePlot(1); break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::TimerTimer(TObject *Sender)
{
	static int n=0;
	static gtime_t t={0};
	gtime_t time={0};
	double rr[6]={0},qr[9]={0},rb[3]={0},age=0.0,ratio=0.0;
	int stat,ns=0;
	
	if ((stat=rtksvrgetsol(&rtksvr,&time,rr,qr,rb,&ns,&age,&ratio))) {
		if (timediff(time,t)!=0.0) {
			UpdateTime();
			UpdatePos();
			UpdateLog(stat,time,rr,qr,rb,ns,age,ratio);
			if (PlotType>=4) UpdatePlot();
			t=time;
		}
		Svr->Color=clLime;
		LabelTime->Font->Color=clBlack;
	}
	else {
		if (PlotType>=4) DrawTrkStat(0);
		IndSol->Color=clWhite;
		Svr->Color=rtksvr.state?clGreen:clBtnFace;
	}
	if (!(++n%20)&&PlotType<4) UpdatePlot();
	if (!(n%20)) UpdateStr();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ChangePlot(int dir)
{
	if (dir) PlotType++; else PlotType--;
	if (PlotType>5) PlotType=0; else if (PlotType<0) PlotType=5;
	PanelFull->Visible=PlotType==5;
	UdScale->Visible=PlotType>=4;
	Panel21->Visible=PlotType!=5;
	Panel22->Visible=PlotType!=5;
	UpdatePlot();
	UpdatePos();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ExpandPlot(int dir)
{
	if (dir&&Scale<18) Scale++; else if (!dir&&Scale>0) Scale--;
	UpdatePlot();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateTimeSys(void)
{
	AnsiString label[]={"GPST","UTC","JST","GPST"};
	BtnTimeSys->Caption=label[TimeSys];
	UpdateTime();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateSolType(void)
{
	AnsiString label[]={
		"Lat/Lon/Height","Lat/Lon/Height","X/Y/Z-ECEF","E/N/U-Baseline",
		"Pitch/Yaw/Length-Baseline"
	};
	Plabel0->Caption=label[SolType];
	UpdatePos();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateLog(int stat, gtime_t time, double *rr,
	double *qr, double *rb, int ns, double age, double ratio)
{
	int ena;
	SolStat[PSolE]=stat; Time[PSolE]=time; Nvsat[PSolE]=ns; Age[PSolE]=age;
	Ratio[PSolE]=ratio;
	for (int i=0;i<3;i++) {
		SolRov[i+PSolE*3]=rr[i];
		SolRef[i+PSolE*3]=rb[i];
		VelRov[i+PSolE*3]=rr[i+3];
	}
	for (int i=0;i<9;i++) Qr[i+PSolE*9]=qr[i];
	PSol=PSolE;
	if (++PSolE>=SolBuffSize) PSolE=0;
	if (PSolE==PSolS&&++PSolS>=SolBuffSize) PSolS=0;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateFont(void)
{
	TLabel *label[]={
		PlabelA,Plabel1,Plabel2,Plabel3,Pos1,Pos2,Pos3,Solution,LabelStd,LabelNSat
	};
	TColor color=label[7]->Font->Color;
	for (int i=0;i<10;i++) label[i]->Font->Assign(PosFont);
	label[0]->Font->Size=9; label[7]->Font->Color=color;
	label[8]->Font->Size=8; label[8]->Font->Color=clGray;
	label[9]->Font->Size=8; label[9]->Font->Color=clGray;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateTime(void)
{
	gtime_t time=Time[PSol];
	double tow;
	int week;
	char tstr[64];
	
	if      (TimeSys==0) time2str(time,tstr,1);
	else if	(TimeSys==1) time2str(gpst2utc(time),tstr,1);
	else if (TimeSys==2) time2str(timeadd(gpst2utc(time),9*3600.0),tstr,1);
	else if	(TimeSys==3) {
		tow=time2gpst(time,&week); sprintf(tstr,"week %04d %8.1f s",week,tow);
	}
	LabelTime->Caption=tstr;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdatePos(void)
{
	TLabel *label[]={Plabel1,Plabel2,Plabel3,Pos1,Pos2,Pos3,LabelStd,LabelNSat};
	AnsiString sol[]={"----","FIX","FLOAT","SBAS","DGPS","SINGLE"},s[8];
	TColor color[]={clSilver,clGreen,CLORANGE,clFuchsia,clBlue,clRed};
	gtime_t time;
	double *rr=SolRov+PSol*3,*rb=SolRef+PSol*3,*qr=Qr+PSol*9,pos[3]={0},Qe[9]={0};
	double dms1[3]={0},dms2[3]={0},bl[3]={0},enu[3]={0},pitch=0.0,yaw=0.0,len;
	int stat=SolStat[PSol];
	
	Solution->Caption=sol[stat];
	Solution->Font->Color=color[stat];
	IndSol->Color=rtksvr.state&&stat?color[stat]:clWhite;
	if (norm(rr,3)>0.0&&norm(rb,3)>0.0) {
		for (int i=0;i<3;i++) bl[i]=rr[i]-rb[i];
	}
	len=norm(bl,3);
	if (SolType==0) {
		if (norm(rr,3)>0.0) {
			ecef2pos(rr,pos); covenu(pos,qr,Qe);
			degtodms(pos[0]*R2D,dms1);
			degtodms(pos[1]*R2D,dms2);
			if (SolOpt.height==1) pos[2]-=geoidh(pos); /* geodetic */
		}
		s[0]=pos[0]<0?"S:":"N:"; s[1]=pos[1]<0?"W:":"E:"; s[2]="H:";
		s[3].sprintf("%.0f%s %02.0f' %07.4f\"",fabs(dms1[0]),CHARDEG,dms1[1],dms1[2]);
		s[4].sprintf("%.0f%s %02.0f' %07.4f\"",fabs(dms2[0]),CHARDEG,dms2[1],dms2[2]);
		s[5].sprintf("%.3f m",pos[2]);
		s[6].sprintf("N:%6.3f E:%6.3f U:%6.3f m",SQRT(Qe[4]),SQRT(Qe[0]),SQRT(Qe[8]));
	}
	else if (SolType==1) {
		if (norm(rr,3)>0.0) {
			ecef2pos(rr,pos); covenu(pos,qr,Qe);
			if (SolOpt.height==1) pos[2]-=geoidh(pos); /* geodetic */
		}
		s[0]=pos[0]<0?"S:":"N:"; s[1]=pos[1]<0?"W:":"E:"; s[2]="H:";
		s[3].sprintf("%.8f %s",fabs(pos[0])*R2D,CHARDEG);
		s[4].sprintf("%.8f %s",fabs(pos[1])*R2D,CHARDEG);
		s[5].sprintf("%.3f m",pos[2]);
		s[6].sprintf("E:%6.3f N:%6.3f U:%6.3f m",SQRT(Qe[0]),SQRT(Qe[4]),SQRT(Qe[8]));
	}
	else if (SolType==2) {
		s[0]="X:"; s[1]="Y:"; s[2]="Z:";
		s[3].sprintf("%.3f m",rr[0]);
		s[4].sprintf("%.3f m",rr[1]);
		s[5].sprintf("%.3f m",rr[2]);
		s[6].sprintf("X:%6.3f Y:%6.3f Z:%6.3f m",SQRT(qr[0]),SQRT(qr[4]),SQRT(qr[8]));
	}
	else if (SolType==3) {
		if (len>0.0) {
			ecef2pos(rb,pos); ecef2enu(pos,bl,enu); covenu(pos,qr,Qe);
		}
		s[0]="E:"; s[1]="N:"; s[2]="U:";
		s[3].sprintf("%.3f m",enu[0]);
		s[4].sprintf("%.3f m",enu[1]);
		s[5].sprintf("%.3f m",enu[2]);
		s[6].sprintf("E:%6.3f N:%6.3f U:%6.3f m",SQRT(Qe[0]),SQRT(Qe[4]),SQRT(Qe[8]));
	}
	else {
		if (len>0.0) {
			ecef2pos(rb,pos); ecef2enu(pos,bl,enu); covenu(pos,qr,Qe);
			pitch=asin(enu[2]/len);
			yaw=atan2(enu[0],enu[1]); if (yaw<0.0) yaw+=2.0*PI;
		}
		s[0]="P:"; s[1]="Y:"; s[2]="L:";
		s[3].sprintf("%.3f %s",pitch*R2D,CHARDEG);
		s[4].sprintf("%.3f %s",yaw*R2D,CHARDEG);
		s[5].sprintf("%.3f m",len);
		s[6].sprintf("E:%6.3f N:%6.3f U:%6.3f m",SQRT(Qe[0]),SQRT(Qe[4]),SQRT(Qe[8]));
	}
	s[7].sprintf("Age:%4.1f s Ratio:%4.1f # of Sat:%2d",Age[PSol],Ratio[PSol],Nvsat[PSol]);
	
	for (int i=0;i<8;i++) label[i]->Caption=s[i];
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateStr(void)
{
	TColor color[]={clRed,clBtnFace,CLORANGE,clGreen,clLime};
	TPanel *ind[MAXSTRRTK]={Str1,Str2,Str3,Str4,Str5,Str6};
	int sstat[MAXSTRRTK]={0};
	char msg[MAXSTRMSG]="";
	
	rtksvrsstat(&rtksvr,sstat,msg);
	for (int i=0;i<MAXSTRRTK;i++) ind[i]->Color=color[sstat[i]+1];
	if (*msg) Message->Caption=msg;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdatePlot(void)
{
	gtime_t time={0};
	TCanvas *c;
	TRect r;
	if (PlotType==5) {
		c=PlotFull->Canvas;
		r=PlotFull->ClientRect;
	}
	else {
		c=Plot->Canvas;
		r=Plot->ClientRect;
	}
	Nsat=rtksvrostat(&rtksvr,PlotType%2,&time,Sat,Az,El,Snr1,Snr2,Vsat);
	c->Brush->Style=bsSolid;
	c->Brush->Color=clWhite;
	c->FillRect(r);
	if		(PlotType<2) DrawSnr();
	else if (PlotType<4) DrawSat();
	else				 DrawTrk();
	if (PlotType==5) DispFull->Canvas->CopyRect(r,c,r);
	else Disp->Canvas->CopyRect(r,c,r);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawSnr(void)
{
	TColor color[]={clGreen,CLORANGE,clFuchsia,clBlue,clRed,clSilver};
	TCanvas *c=Plot->Canvas;
	AnsiString s; 
	int ns=0;
	for (int i=0;i<Nsat;i++) {
		if (Snr1[i]>0||Snr2[i]>0) ns++;
	}
	int h=Plot->Height-30,w=(Plot->Width)/(ns+1);
	c->Pen->Color=clSilver;
	for (int snr=MINSNR+10;snr<MAXSNR;snr+=10) {
		int y=15+h-(snr-MINSNR)*h/(MAXSNR-MINSNR);
		c->MoveTo(3,y); c->LineTo(Plot->Width-13,y);
		DrawText(c,Plot->Width-9,y,s.sprintf("%d",snr),clGray,1);
	}
	DrawText(c,3,1,PlotType%2?"Base SNR (dBHz)":"Rover SNR (dBHz)",clGray,0);
	c->Pen->Color=clGray;
	c->Brush->Style=bsClear;
	TRect b(1,16,Plot->Width-2,15+h);
	c->Rectangle(b);
	for (int i=0,j=0;i<Nsat&&i<MAXSAT;i++) {
		if (Snr1[i]<=0&&Snr2[i]<=0) continue;
		int x1=2+j*w+2,x2=x1+w-3;
		int y1=15+h;
		int y2=y1-(Snr1[i]-MINSNR)*h/(MAXSNR-MINSNR);
		int y3=y1-(Snr2[i]-MINSNR)*h/(MAXSNR-MINSNR);
		int k1=(49-Snr1[i])/5,k2=(49-Snr2[i])/5;
		y2=y2<2?2:(y1<y2?y1:y2);
		y3=y3<2?2:(y1<y3?y1:y3);
		TRect r1(x1,y1,(x1+x2)/2+1,y2),r2((x1+x2)/2,y1,x2,y3);
		c->Brush->Style=bsSolid;
		if (Snr1[i]>0) {
			c->Brush->Color=color[k1<0?0:(k1>5?5:k1)];
			c->FillRect(r1); c->Rectangle(r1);
		}
		if (Snr2[i]>0) {
			c->Brush->Color=color[k2<0?0:(k2>5?5:k2)];
			c->FillRect(r2); c->Rectangle(r2);
		}
		c->Brush->Style=bsClear;
		AnsiString s; s.sprintf("%d",Sat[i]);
		DrawText(c,x1+w/2-1,y1+6,s,Vsat[i]?clGray:clSilver,1);
		j++;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawSat(void)
{
	TColor color[]={clGreen,CLORANGE,clFuchsia,clBlue,clRed,clSilver};
	TCanvas *c=Plot->Canvas;
	AnsiString s;
	TPoint p(Plot->Width/2,Plot->Height/2);
	double r=Plot->Width*0.85/2,azel[MAXSAT*2],dop[4];
	DrawSky();
	int x[MAXSAT],y[MAXSAT],ns=0;
	for (int i=0;i<Nsat&&i<MAXSAT;i++) {
		if (El[i]<=0.0) continue;
		if (Vsat[i]) {
			azel[ns*2]=Az[i]; azel[1+ns*2]=El[i];
			ns++;
		}
		x[i]=(int)(p.x+r*(90-El[i]*R2D)/90*sin(Az[i]));
		y[i]=(int)(p.y-r*(90-El[i]*R2D)/90*cos(Az[i]));
		c->Pen->Color=clGray;
		c->Brush->Style=bsSolid;
		int j=(49-Snr1[i])/5,d=SATSIZE/2;
		c->Brush->Color=Vsat[i]?color[j<0?0:(j>5?5:j)]:clSilver;
		c->Ellipse(x[i]-d,y[i]-d,x[i]+d,y[i]+d);
		c->Brush->Style=bsClear;
		AnsiString s; s.sprintf("%d",Sat[i]);
		DrawText(c,x[i],y[i],s,clWhite,1);
	}
	c->Brush->Style=bsClear;
	dops(ns,azel,0.0,dop);
	DrawText(c,3,1,PlotType%2?"Base Sats":"Rover Sats",clGray,0);
	DrawText(c,3,Plot->Height-15,s.sprintf("# of Sat:%2d",Nsat),clGray,0);
	DrawText(c,Plot->Width-3,Plot->Height-15,s.sprintf("GDOP:%.1f",dop[0]),clGray,2);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawSky(void)
{
	TCanvas *c=Plot->Canvas;
	TPoint p(Plot->Width/2,Plot->Height/2);
	double r=Plot->Width*0.85/2;
	c->Brush->Color=clWhite;
	c->Brush->Style=bsSolid;
	for (int e=0;e<90;e+=30) {
		int d=(int)(r*(90-e)/90);
		c->Pen->Color=e==0?clGray:clSilver;
		c->Ellipse(p.x-d,p.y-d,p.x+d,p.y+d);
	}
	AnsiString label[]={"N","E","S","W"};
	for (int a=0;a<360;a+=45) {
		int x=(int)(r*sin(a*D2R)),y=(int)(r*cos(a*D2R));
		c->Pen->Color=clSilver;
		c->MoveTo(p.x,p.y); c->LineTo(p.x+x,p.y-y);
		if (a%90==0) DrawText(c,p.x+x,p.y-y,label[a/90],clGray,1);
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawTrk(void)
{
	TColor color[]={clWhite,clGreen,CLORANGE,clFuchsia,clBlue,clRed};
	AnsiString s;
	TCanvas *c;
	TPoint *p,p0,p1,p2,p3[6];
	double pos[3],rr[3],enu0[3],enu[3],enuv[3],*ref,vel,ang;
	double fact,scale[]={
		0.01,0.02,0.05,0.1,0.2,0.5,1,2,5,10,20,50,100,200,500,1000,2000,5000,10000
	};
	int n=0,*stat,w,h;
	int gint[]={20,20,16,20,20,16,20,20,16,20,20,16,20,20,16,20,20,16,20};
	int ng[]={10,5,5,10,5,5,10,5,5,10,5,5,10,5,5,10,5,5,10};
	int pc[][2]={{5,3},{5,11},{5,7},{45,7},{45,3},{45,11}};
	c=PlotType==5?PlotFull->Canvas:Plot->Canvas;
	w=PlotType==5?PlotFull->Width:Plot->Width;
	h=PlotType==5?PlotFull->Height:Plot->Height;
	
	p=new TPoint[SolBuffSize];
	stat=new int[SolBuffSize];
	
	fact=40.0/scale[Scale];
	ref=norm(SolRef+PSol*3,3)<=0.0?SolRov+PSolS*3:SolRef+PSol*3;
	
	ecef2pos(ref,pos);
	for (int i=0;i<3;i++) rr[i]=SolRov[i+PSol*3]-ref[i];
	ecef2enu(pos,rr,enu0);
	ecef2enu(pos,VelRov+PSol*3,enuv);
	
	p0.x=w/2; p0.y=h/2;
	p1.x=p0.x-(int)(enu0[0]*fact+0.5);
	p1.y=p0.y+(int)(enu0[1]*fact+0.5);
	p2.x=VELSIZE/2+5;
	p2.y=h-VELSIZE/2-15;
	
	DrawCircle(c,p2,VELSIZE,CLLGRAY,clWhite);
	DrawGrid(c,w,h,p1,gint[Scale],ng[Scale],CLLGRAY,clSilver);
	DrawCircle(c,p1,SOLSIZE,clSilver,clSilver);
	
	for (int i=PSolS;i!=PSolE;) {
		for (int j=0;j<3;j++) rr[j]=SolRov[j+i*3]-SolRef[j+i*3];
		ecef2enu(pos,rr,enu);
		p[n].x=p0.x+(int)((enu[0]-enu0[0])*fact+0.5);
		p[n].y=p0.y-(int)((enu[1]-enu0[1])*fact+0.5);
		stat[n++]=SolStat[i];
		if (++i>=SolBuffSize) i=0;
	}
	c->Pen->Color=CLLGRAY;
	c->Pen->Style=psSolid;
	c->Polyline(p,n-1);
	for (int i=0;i<n;i++) {
		DrawCircle(c,p[i],HISSIZE,color[stat[i]],color[stat[i]]);
	}
	c->Pen->Color=clBlack;
	c->MoveTo(p0.x-SOLSIZE/2-4,p0.y); c->LineTo(p0.x+SOLSIZE/2+4,p0.y);
	c->MoveTo(p0.x,p0.y-SOLSIZE/2-4); c->LineTo(p0.x,p0.y+SOLSIZE/2+4);
	DrawCircle(c,p0,SOLSIZE+4,clBlack,clWhite);
	DrawCircle(c,p0,SOLSIZE,clBlack,color[SolStat[PSol]]);
	
	c->Brush->Style=bsClear;
	DrawText(c,p2.x,p2.y-VELSIZE/2,"N",clSilver,1);
	if ((vel=norm(enuv,3))>=1.0) {
		ang=atan2(enuv[0],enuv[1])*R2D;
		DrawArrow(c,p2,VELSIZE-4,(int)ang,clGray);
	}
	DrawCircle(c,p2,9,vel<1.0?CLLGRAY:clGray,clWhite);
	c->Brush->Style=bsClear;
	DrawText(c,p2.x,p2.y+VELSIZE/2+7,s.sprintf("%.0f km/h",vel*3.6),clGray,1);
	
	c->Pen->Color=clGray;
	for (int i=0;i<6;i++) {
		p3[i].x=w-30-pc[i][0];
		p3[i].y=h-pc[i][1];
	}
	c->Polyline(p3,5);
	c->Brush->Style=bsClear;
	char *unit="m";
	double sc=scale[Scale];
	if (sc<1.0) {sc*=100.0; unit="cm";}
	else if (sc>=1000.0) {sc/=1000.0; unit="km";}
	DrawText(c,w-55,h-15,s.sprintf("%.0f %s",sc,unit),clGray,1);
	
	c->Brush->Style=bsSolid;
	c->Brush->Color=clWhite;
	if (PlotType==5) {
		DrawText(c,3,1,Solution->Caption,Solution->Font->Color,0);
		s=Plabel1->Caption+Pos1->Caption+" "+
		  Plabel2->Caption+Pos2->Caption+" "+
		  Plabel3->Caption+Pos3->Caption+" ";
		DrawText(c,45,1,s,clBlack,0);
		DrawText(c,w/2,h-8,LabelNSat->Caption,clGray,1);
	}
	else {
		DrawText(c,3,1,"Gnd Track",clGray,0);
	}
	delete [] p; delete [] stat;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawTrkStat(int stat)
{
	TColor color[]={clWhite,clGreen,CLORANGE,clFuchsia,clBlue,clRed};
	TCanvas *c;
	TPoint p0;
	TRect r;
	int w,h;
	c=PlotType==5?PlotFull->Canvas:Plot->Canvas;
	r=PlotType==5?PlotFull->ClientRect:Plot->ClientRect;
	w=PlotType==5?PlotFull->Width:Plot->Width;
	h=PlotType==5?PlotFull->Height:Plot->Height;
	p0.x=w/2; p0.y=h/2;
	DrawCircle(c,p0,SOLSIZE,clBlack,color[stat]);
	if (PlotType==5) DispFull->Canvas->CopyRect(r,c,r);
	else Disp->Canvas->CopyRect(r,c,r);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawCircle(TCanvas *c, TPoint p, int r, TColor color1,
	TColor color2)
{
	int x1=p.x-r/2,x2=x1+r,y1=p.y-r/2,y2=y1+r;
	c->Pen->Color=color1; c->Pen->Style=psSolid; c->Pen->Width=1;
	c->Brush->Color=color2; c->Brush->Style=bsSolid;
	c->Ellipse(x1,y1,x2,y2);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawGrid(TCanvas *c, int w, int h, TPoint p,
	int gint, int ng, TColor color1, TColor color2)
{
	int i,x,y;
	for (i=-p.x/gint;i<=(w-p.x)/gint;i++) {
		c->Pen->Color=i%ng?color1:color2;
		c->MoveTo(p.x+i*gint,0); c->LineTo(p.x+i*gint,h);
	}
	for (i=-p.y/gint;i<=(h-p.y)/gint;i++) {
		c->Pen->Color=i%ng?color1:color2;
		c->MoveTo(0,p.y+i*gint); c->LineTo(w,p.y+i*gint);
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawText(TCanvas *c, int x, int y, AnsiString s,
	TColor color, int align)
{
	TSize off=c->TextExtent(s);
	c->Font->Charset=ANSI_CHARSET;
	if (align==1) {x-=off.cx/2; y-=off.cy/2;} else if (align==2) x-=off.cx;
	c->Font->Color=color;
	c->TextOut(x,y,s);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawArrow(TCanvas *c, TPoint p, int siz, int ang,
	TColor color)
{
	TPoint p1[6],p2[6];
	p1[0].x=p1[1].x=0; p1[2].x=3; p1[3].x=0; p1[4].x=-3; p1[5].x=0;
	p1[0].y=-siz/2; p1[1].y=siz/2; p1[2].y=p1[4].y=p1[1].y-6;
	p1[3].y=p1[5].y=siz/2;
	
	for (int i=0;i<6;i++) {
		p2[i].x=p.x+(int)(p1[i].x*cos(-ang*D2R)-p1[i].y*sin(-ang*D2R)+0.5);
		p2[i].y=p.y-(int)(p1[i].x*sin(-ang*D2R)+p1[i].y*cos(-ang*D2R)+0.5);
	}
	c->Pen->Color=color;
	c->Polyline(p2,5);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::InitSolBuff(void)
{
	double ep[]={2000,1,1,0,0,0};
	
	delete [] Time;   delete [] SolStat; delete [] Nvsat;  delete [] SolRov;
	delete [] SolRef; delete [] Qr;      delete [] VelRov; delete [] Age;
	delete [] Ratio;
	
	if (SolBuffSize<=0) SolBuffSize=1;
	Time   =new gtime_t[SolBuffSize];
	SolStat=new int[SolBuffSize];
	Nvsat  =new int[SolBuffSize];
	SolRov =new double[SolBuffSize*3];
	SolRef =new double[SolBuffSize*3];
	VelRov =new double[SolBuffSize*3];
	Qr     =new double[SolBuffSize*9];
	Age    =new double[SolBuffSize];
	Ratio  =new double[SolBuffSize];
	PSol=PSolS=PSolE=0;
	for (int i=0;i<SolBuffSize;i++) {
		Time[i]=epoch2time(ep);
		SolStat[i]=Nvsat[i]=0;
		for (int j=0;j<3;j++) SolRov[j+i*3]=SolRef[j+i*3]=VelRov[j+i*3]=0.0;
		for (int j=0;j<9;j++) Qr[j+i*9]=0.0;
		Age[i]=Ratio[i]=0.0;
	}
	ScbSol->Max=0; ScbSol->Position=0;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SaveLog(void)
{
	FILE *fp;
	int posf[]={SOLF_LLH,SOLF_LLH,SOLF_XYZ,SOLF_ENU,SOLF_ENU};
	solopt_t opt;
	double  ep[6],pos[3];
	char file[1024];
	
	time2epoch(timeget(),ep);
	sprintf(file,"rtk_%04.0f%02.0f%02.0f%02.0f%02.0f%02.0f.txt",
			ep[0],ep[1],ep[2],ep[3],ep[4],ep[5]);
	SaveDialog->FileName=file;
	if (!SaveDialog->Execute()) return;
	if (!(fp=fopen(SaveDialog->FileName.c_str(),"wt"))) {
		Message->Caption="log file open error"; return;
	}
	opt=SolOpt;
	opt.posf=posf[SolType];
	setsolopt(&opt);
	if (SolOpt.outhead) {
		fprintf(fp,"%% program   : %s ver.%s\n",PRGNAME,VER_RTKLIB);
		if (PrcOpt.mode==PMODE_DGPS||PrcOpt.mode==PMODE_KINEMA||
			PrcOpt.mode==PMODE_STATIC) {
			ecef2pos(PrcOpt.rb,pos);
			fprintf(fp,"%% ref pos   :%13.9f %14.9f %10.4f\n",pos[0]*R2D,
					pos[1]*R2D,pos[2]);
		}
		fprintf(fp,"%%\n");
	}
	outsolhead(fp);
	int s,n=PSolE-PSolS;
	if (n<0) n+=SolBuffSize;
	if (n<=SavedSol) s=PSolS; else s=PSolE-SavedSol;
	if (s<0) s+=SolBuffSize;
	for (int i=s;i!=PSolE;) {
		outsol(fp,Time[i],SolRov+i*3,Qr+i*9,SolRef+i*3,SolStat[i],Nvsat[i]);
		if (++i>=SolBuffSize) i=0;
	}
	fclose(fp);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LoadNav(nav_t *nav)
{
	TIniFile *ini=new TIniFile("rtknavi.ini");
	AnsiString str,s;
	eph_t eph0={0};
	
	for (int i=0;i<MAXSAT;i++) {
		if ((str=ini->ReadString("navi",s.sprintf("eph_%02d",i),""))=="") continue;
		nav->eph[i]=eph0;
		sscanf(str.c_str(),
			   "%d,%d,%d,%d,%d,%d,%d,%d,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf",
			   &nav->eph[i].sat,
			   &nav->eph[i].iode,
			   &nav->eph[i].iodc,
			   &nav->eph[i].sva ,
			   &nav->eph[i].svh ,
			   &nav->eph[i].toe.time,
			   &nav->eph[i].toc.time,
			   &nav->eph[i].ttr.time,
			   &nav->eph[i].A   ,
			   &nav->eph[i].e   ,
			   &nav->eph[i].i0  ,
			   &nav->eph[i].OMG0,
			   &nav->eph[i].omg ,
			   &nav->eph[i].M0  ,
			   &nav->eph[i].deln,
			   &nav->eph[i].OMGd,
			   &nav->eph[i].idot,
			   &nav->eph[i].crc ,
			   &nav->eph[i].crs ,
			   &nav->eph[i].cuc ,
			   &nav->eph[i].cus ,
			   &nav->eph[i].cic ,
			   &nav->eph[i].cis ,
			   &nav->eph[i].toes,
			   &nav->eph[i].fit ,
			   &nav->eph[i].f0  ,
			   &nav->eph[i].f1  ,
			   &nav->eph[i].f2  ,
			   &nav->eph[i].tgd);
	}
	str=ini->ReadString("navi","ion","");
	for (int i=0;i<8;i++) nav->ion[i]=0.0;
	sscanf(str.c_str(),"%f,%f,%f,%f,%f,%f,%f,%f",nav->ion,nav->ion+1,nav->ion+2,
		   nav->ion+3,nav->ion+4,nav->ion+5,nav->ion+6,nav->ion+7);
	str=ini->ReadString("navi","utc","");
	
	for (int i=0;i<4;i++) nav->utc[i]=0.0;
	sscanf(str.c_str(),"%f,%f,%f,%f",nav->utc,nav->utc+1,nav->utc+2,nav->utc+3);
	
	nav->leaps=ini->ReadInteger("navi","leaps",0);
	delete ini;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SaveNav(nav_t *nav)
{
	TIniFile *ini=new TIniFile("rtknavi.ini");
	AnsiString str,s;
	
	for (int i=0;i<MAXSAT;i++) {
		if (nav->eph[i].ttr.time==0) continue;
		str="";
		str=str+s.sprintf("%d,",nav->eph[i].sat);
		str=str+s.sprintf("%d,",nav->eph[i].iode);
		str=str+s.sprintf("%d,",nav->eph[i].iodc);
		str=str+s.sprintf("%d,",nav->eph[i].sva);
		str=str+s.sprintf("%d,",nav->eph[i].svh);
		str=str+s.sprintf("%d,",(int)nav->eph[i].toe.time);
		str=str+s.sprintf("%d,",(int)nav->eph[i].toc.time);
		str=str+s.sprintf("%d,",(int)nav->eph[i].ttr.time);
		str=str+s.sprintf("%.14E,",nav->eph[i].A);
		str=str+s.sprintf("%.14E,",nav->eph[i].e);
		str=str+s.sprintf("%.14E,",nav->eph[i].i0);
		str=str+s.sprintf("%.14E,",nav->eph[i].OMG0);
		str=str+s.sprintf("%.14E,",nav->eph[i].omg);
		str=str+s.sprintf("%.14E,",nav->eph[i].M0);
		str=str+s.sprintf("%.14E,",nav->eph[i].deln);
		str=str+s.sprintf("%.14E,",nav->eph[i].OMGd);
		str=str+s.sprintf("%.14E,",nav->eph[i].idot);
		str=str+s.sprintf("%.14E,",nav->eph[i].crc);
		str=str+s.sprintf("%.14E,",nav->eph[i].crs);
		str=str+s.sprintf("%.14E,",nav->eph[i].cuc);
		str=str+s.sprintf("%.14E,",nav->eph[i].cus);
		str=str+s.sprintf("%.14E,",nav->eph[i].cic);
		str=str+s.sprintf("%.14E,",nav->eph[i].cis);
		str=str+s.sprintf("%.14E,",nav->eph[i].toes);
		str=str+s.sprintf("%.14E,",nav->eph[i].fit);
		str=str+s.sprintf("%.14E,",nav->eph[i].f0);
		str=str+s.sprintf("%.14E,",nav->eph[i].f1);
		str=str+s.sprintf("%.14E,",nav->eph[i].f2);
		str=str+s.sprintf("%.14E,",nav->eph[i].tgd);
		ini->WriteString("navi",s.sprintf("eph_%02d",i),str);
	}
	str="";
	for (int i=0;i<8;i++) str=str+s.sprintf("%.14E,",nav->ion[i]);
	ini->WriteString("navi","ion",str);
	
	str="";
	for (int i=0;i<4;i++) str=str+s.sprintf("%.14E,",nav->utc[i]);
	ini->WriteString("navi","utc",str);
	
	ini->WriteInteger("navi","leaps",nav->leaps);
	delete ini;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BtnAboutClick(TObject *Sender)
{
	AboutDialog->About=PRGNAME;
	AboutDialog->IconIndex=5;
	AboutDialog->ShowModal();
}
//---------------------------------------------------------------------------

